@@ -5,6 +5,7 @@ import functools
 import polars as pl
 from collections import defaultdict
 
+import os
 import shutil
 import numpy as np
 
@@ -12,7 +13,7 @@ import torch
 from torch.utils.data import DataLoader
 from torch.utils.data import Dataset
 
-from tf_bind_transformer.protein_utils import parse_gene_name
+from tf_bind_transformer.gene_utils import parse_gene_name
 from enformer_pytorch import FastaInterval
 
 from pyfaidx import Fasta
@@ -29,6 +30,14 @@ def cast_list(val = None):
 def read_bed(path):
     return pl.read_csv(path, sep = '\t', has_headers = False)
 
+def save_bed(df, path):
+    df.to_csv(path, sep = '\t', has_header = False)
+
+def parse_exp_target_cell(exp_target_cell):
+    experiment, target, *cell_type = exp_target_cell.split('.')
+    cell_type = '.'.join(cell_type) # handle edge case where cell type contains periods
+    return experiment, target, cell_type
+
 # fetch protein sequences by gene name and uniprot id
 
 class FactorProteinDatasetByUniprotID(Dataset):
@@ -187,6 +196,37 @@ def generate_random_ranges_from_fasta(
         filter_bed_file_by_(tmp_file, file, tmp_file)
 
     shutil.move(tmp_file, f'./{output_filename}')
+
+    print('success')
+
+def generate_negative_peaks_per_target(
+    remap_file,
+    *,
+    output_folder = './negative-peaks-per-target',
+    exp_target_cell_type_col = 'column_4',
+    id_column = 'column_5'
+):
+    output_folder = Path(output_folder)
+    output_folder.mkdir(exist_ok = True, parents = True)
+
+    df = read_bed(remap_file)
+    target_experiments = df.get_column(exp_target_cell_type_col).unique().to_list()
+
+    for target_experiment in target_experiments:
+        filtered_df = df.filter(pl.col(exp_target_cell_type_col) == target_experiment)
+
+        target_bed_path = str(output_folder / f'{target_experiment}.bed')
+        negative_peak_path = str(output_folder / f'{target_experiment}.negative.bed')
+
+        save_bed(filtered_df, target_bed_path)
+        filter_bed_file_by_(remap_file, target_bed_path, negative_peak_path)
+        os.remove(target_bed_path)
+
+        neg_df = read_bed(negative_peak_path)
+        np_array = neg_df.get_column(id_column).to_numpy()
+        np.save(str(output_folder / target_experiment), np_array)
+        os.remove(negative_peak_path)
+
     print('success')
 
 # dataset for remap data - all peaks
@@ -260,8 +300,7 @@ class RemapAllPeakDataset(Dataset):
     def __getitem__(self, ind):
         chr_name, begin, end, _, _, _, experiment_target_cell_type, reading, *_ = self.df.row(ind)
 
-        experiment, target, *cell_type = experiment_target_cell_type.split('.')
-        cell_type = '.'.join(cell_type) # handle edge case where cell type contains periods
+        experiment, target, cell_type = parse_exp_target_cell(experiment_target_cell_type)
 
         seq = self.fasta(chr_name, begin, end)
         aa_seq = self.factor_ds[target]
@@ -286,6 +325,7 @@ class NegativePeakDataset(Dataset):
         include_targets = None,
         exclude_cell_types = None,
         include_cell_types = None,
+        exp_target_cell_column = 'column_4',
         **kwargs
     ):
         super().__init__()
@@ -310,7 +350,6 @@ class NegativePeakDataset(Dataset):
         if exists(filter_chromosome_ids):
             dataset_chr_ids = dataset_chr_ids.intersection(set(filter_chromosome_ids))
 
-
         neg_df = neg_df.filter(pl_isin('column_1', get_chr_names(dataset_chr_ids)))
 
         assert len(neg_df) > 0, 'dataset is empty by filter criteria'
@@ -342,6 +381,16 @@ class NegativePeakDataset(Dataset):
         if exclude_cell_types:
             self.cell_types = list(set(self.cell_types) - set(exclude_cell_types))
 
+        # get all exp-target-cells and filter by above
+        exp_target_cells = remap_df.get_column(exp_target_cell_column).unique().to_list()
+
+        def filter_exp_target_cells(el):
+            experiment, target, cell_type = parse_exp_target_cell(el)
+            return experiment in self.experiments and target in self.targets and cell_type in self.cell_types
+
+        self.filtered_exp_target_cells = list(filter(filter_exp_target_cells, exp_target_cells))
+        assert len(self.filtered_exp_target_cells), 'no experiment-target-cell left for hard negative set'
+
         self.factor_ds = FactorProteinDataset(factor_fasta_folder)
         self.fasta = FastaInterval(**kwargs)
 
@@ -351,9 +400,8 @@ class NegativePeakDataset(Dataset):
     def __getitem__(self, ind):
         chr_name, begin, end = self.neg_df.row(ind)
 
-        experiment = choice(self.experiments)
-        target = choice(self.targets)
-        cell_type = choice(self.cell_types)
+        random_exp_target_cell_type = choice(self.filtered_exp_target_cells)
+        experiment, target, cell_type = parse_exp_target_cell(random_exp_target_cell_type)
 
         seq = self.fasta(chr_name, begin, end)
         aa_seq = self.factor_ds[target]
