@@ -1,3 +1,4 @@
+import copy
 import math
 import torch
 import torch.nn.functional as F
@@ -12,6 +13,8 @@ from enformer_pytorch import Enformer
 from enformer_pytorch.enformer_pytorch import poisson_loss, pearson_corr_coef
 from enformer_pytorch.finetune import freeze_batchnorms_, freeze_all_but_layernorms_
 
+from logavgexp_pytorch import logavgexp
+
 from tf_bind_transformer.cache_utils import cache_fn
 from tf_bind_transformer.protein_utils import get_esm_repr, ESM_EMBED_DIM
 from tf_bind_transformer.context_utils import get_text_repr, get_contextual_dim
@@ -31,35 +34,9 @@ def null_context():
 
 # tensor helpers
 
-def log(t, eps = 1e-20):
-    return torch.log(t + eps)
-
 def l2norm(t):
     return F.normalize(t, dim = -1)
 
-def logavgexp(
-    t,
-    mask = None,
-    dim = -1,
-    eps = 1e-20,
-    temp = 0.01
-):
-    if exists(mask):
-        mask_value = -torch.finfo(t.dtype).max
-        t = t.masked_fill(~mask, mask_value)
-        n = mask.sum(dim = dim)
-        norm = torch.log(n)
-    else:
-        n = t.shape[dim]
-        norm = math.log(n)
-
-    t = t / temp
-    max_t = t.amax(dim = dim).detach()
-    t_exp = (t - max_t.unsqueeze(dim)).exp()
-    avg_exp = t_exp.sum(dim = dim).clamp(min = eps) / n
-    out = log(avg_exp, eps = eps) + max_t - norm
-    return out * temp
-
 # genetic sequence caching enformer forward decorator
 
 def cache_enformer_forward(fn):
@@ -341,7 +318,8 @@ class AttentionAdapterModel(nn.Module):
         dropout = 0.,
         use_squeeze_excite = False,
         binary_target = False,
-        target_mse_loss = False
+        target_mse_loss = False,
+        read_value_aux_loss_weight = 0.1
     ):
         super().__init__()
         assert isinstance(enformer, Enformer), 'enformer must be an instance of Enformer'
@@ -407,6 +385,8 @@ class AttentionAdapterModel(nn.Module):
                 nn.Linear(enformer_dim, 1),
                 Rearrange('... 1 -> ...')
             )
+
+            self.to_read_value_pred = copy.deepcopy(self.to_pred)
         else:
             self.to_pred = nn.Sequential(
                 nn.Linear(enformer_dim, 1),
@@ -424,6 +404,7 @@ class AttentionAdapterModel(nn.Module):
         contextual_free_text = None,
         aa_mask = None,
         target = None,
+        read_value = None,
         return_corr_coef = False,
         finetune_enformer = False,
         finetune_enformer_ln_only = False
@@ -519,17 +500,24 @@ class AttentionAdapterModel(nn.Module):
 
         pred = self.to_pred(logits)
 
-        if exists(target):
-            if self.binary_target:
-                return self.loss_fn(pred, target.float())
-            else:
-                if return_corr_coef:
-                    return pearson_corr_coef(pred, target)
+        if not exists(target):
+            return pred
 
-                return poisson_loss(pred, target)
+        if exists(target) and return_corr_coef:
+            return pearson_corr_coef(pred, target)
+
+        if self.binary_target:
+            return self.loss_fn(pred, target.float())
+        else:
+            return poisson_loss(pred, target)
 
         # return prediction if not auto-calculating loss
 
+        if exists(read_value):
+            read_value_pred = self.to_read_value_pred(seq_embed_convolved_read)
+            read_value = rearrange(read_value, 'b 1 -> b')
+            loss = loss + F.smooth_l1_loss(read_value_pred, read_value) * self.read_value_aux_loss_weight
+
         return pred
 
 # contextual transformer producing weights of convolution on genetic embeddings
